// 定义一个类
class chatAIUtils {
  // 构造器方法，用于初始化新创建的对象
  constructor(name) {
    this.name = name;
  }
  /**
   * @type {{version:String,type:String}}
   * @static
   */
  static version
  static currentNoteId
  static currentNotebookId
  static currentSelection = ""
  static onPopupMenuOnNoteTime = 0
  static onPopupMenuOnSelectionTime = 0
  static currentPrompt
  /**
   * @type {Boolean}
   * @static
   */
  static OCREnhancedMode
  static lastTime
  /**
   * @type {chatglmController}
   * @static
   */
  static chatController
  /**
   * @type {notificationController}
   * @static
   */
  static notifyController
  /**
   * @type {dynamicController}
   * @static
   */
  static dynamicController
  static init(){
    this.OCREnhancedMode = false
    this.app = Application.sharedInstance()
    this.data = Database.sharedInstance()
    this.focusWindow = this.app.focusWindow
    this.version = this.appVersion()
  }
  static showHUD(message,duration=2) {
    this.app.showHUD(message,this.focusWindow,2)
  }

  static appVersion() {
    let info = {}
    let version = parseFloat(this,this.app.appVersion)
    if (version >= 4) {
      info.version = "marginnote4"
    }else{
      info.version = "marginnote3"
    }
    switch (this.app.osType) {
      case 0:
        info.type = "iPadOS"
        break;
      case 1:
        info.type = "iPhoneOS"
        break;
      case 2:
        info.type = "macOS"
        break;
      default:
        break;
    }
    return info
  }
  static  getNoteColors() {
    return ["#ffffb4","#ccfdc4","#b4d1fb","#f3aebe","#ffff54","#75fb4c","#55bbf9","#ea3323","#ef8733","#377e47","#173dac","#be3223","#ffffff","#dadada","#b4b4b4","#bd9fdc"]
  }
  static getNoteById(noteid) {
    let note = this.data.getNoteById(noteid)
    return note
  }
  static getNoteBookById(notebookId) {
    let notebook = this.data.getNotebookById(notebookId)
    return notebook
  }
  static getUrlByNoteId(noteid) {
    let ver = this.appVersion()
    return ver.version+'app://note/'+noteid
  }
  /**
   * 
   * @param {String} url 
   * @returns {String}
   */
  static getNoteIdByURL(url) {
    let targetNoteId = url.trim()
    if (/^marginnote\dapp:\/\/note\//.test(targetNoteId)) {
      targetNoteId = targetNoteId.slice(22)
    }
    return targetNoteId
  }
  static clipboardText() {
    return UIPasteboard.generalPasteboard().string
  }
  static copy(text) {
    UIPasteboard.generalPasteboard().string = text
  }
  static copyJSON(object) {
    UIPasteboard.generalPasteboard().string = JSON.stringify(object,null,2)
  }
  /**
   * 
   * @param {NSData} imageData 
   */
  static copyImage(imageData) {
    UIPasteboard.generalPasteboard().setDataForPasteboardType(imageData,"public.png")
  }
  static studyController() {
    return this.app.studyController(this.focusWindow)
  }
  static studyView() {
    return this.app.studyController(this.focusWindow).view
  }
  
  static currentDocController() {
    return this.studyController().readerController.currentDocumentController
  }
  static currentNotebook() {
    let notebookId = this.studyController().notebookController.notebookId
    return this.getNoteBookById(notebookId)
  }
  static undoGrouping(notebookId,f){
    UndoManager.sharedInstance().undoGrouping(
      String(Date.now()),
      notebookId,
      f
    )
    this.app.refreshAfterDBChanged(notebookId)
  }
  /**
   * 
   * @param {{title:String,content:String,markdown:Boolean,color:Number}} config 
   * @returns 
   */
  static creatNote(config) {
    try {

    let notebook = this.currentNotebook()
    let title = config.title ?? ""
    let note = Note.createWithTitleNotebookDocument(title, notebook,this.currentDocController().document)
    if (config.content) {
      if (config.markdown) {
        note.appendMarkdownComment(config.content)
      }else{
        note.appendTextComment(config.content)
      }
    }
    if (config.color !== undefined) {
      note.colorIndex = config.color
    }
    return note
    } catch (error) {
      this.showHUD(error)
      return undefined
    }
  }
  /**
   * 
   * @param {MbBookNote} parent 
   * @param {MbBookNote} targetNote 
   */
  static addChildNote(parent,targetNote,colorInheritance=false) {
    this.undoGrouping(parent.notebookId, ()=>{
      if (colorInheritance) {
        targetNote.colorIndex = parent.colorIndex
      }
      parent.addChild(targetNote)
    })
  }
  /**
   * 
   * @param {MbBookNote|String} parent 
   * @param {{title:String,content:String,markdown:Boolean,color:Number}} config 
   * @returns {MbBookNote}
   */
  static createChildNote(parent,config) {
    let parentNote = (typeof parent === "string")?this.getNoteById(parent):parent
    let child
    this.undoGrouping(parentNote.notebookId, ()=>{
      child = this.creatNote(config)
      parentNote.addChild(child)
    })
    return child
  }
  /**
   * 
   * @param {MbBookNote} brother 
   * @param {MbBookNote} targetNote 
   */
  static addBrotherNote(brother,targetNote,colorInheritance=false) {
    if (!brother.parentNote) {
      this.showHUD("No parent note!")
      return
    }
    let parent = brother.parentNote
    this.undoGrouping(parent.notebookId, ()=>{
      if (colorInheritance) {
        targetNote.colorIndex = brother.colorIndex
      }
      parent.addChild(targetNote)
    })
  }
  /**
   * 
   * @param {MbBookNote} brother 
   * @param {{title:String,content:String,markdown:Boolean,color:Number}} config 
   */
  static createBrotherNote(brother,config) {
    if (!brother.parentNote) {
      this.showHUD("No parent note!")
      return
    }
    let parent = brother.parentNote
    this.undoGrouping(parent.notebookId, ()=>{
      let child = this.creatNote(config)
      parent.addChild(child)
    })
  }
  static getSplitLine() {
    let study = this.studyController()
    let studyFrame = study.view.bounds
    let readerFrame = study.readerController.view.frame
    let hidden = study.readerController.view.hidden//true代表脑图全屏
    let rightMode = study.rightMapMode
    let fullWindow = readerFrame.width == studyFrame.width
    if (hidden || fullWindow) {
      return undefined
    }
    if (rightMode) {
      let splitLine = readerFrame.x+readerFrame.width
      return splitLine
    }else{
      let splitLine = readerFrame.x
      return splitLine
    }
  }
  static hexColorAlpha(hex,alpha) {
    let color = UIColor.colorWithHexString(hex)
    return alpha!==undefined?color.colorWithAlphaComponent(alpha):color
  }
  static genFrame(x,y,width,height) {
    return {x:x,y:y,width:width,height:height}
  }
  static getImage(path,scale=2) {
    return UIImage.imageWithDataScale(NSData.dataWithContentsOfFile(path), scale)
  }

  static getFocusNote() {
    let notebookController = this.studyController().notebookController
    // copy("text:"+!notebookController.view.hidden)
    if (!notebookController.view.hidden && notebookController.mindmapView && notebookController.focusNote) {
      return notebookController.focusNote
    }else{
      return (this.currentDocController().focusNote ?? this.currentDocController().lastFocusNote)
    }
  }

  static getFocusNotes() {
    let notebookController = this.studyController().notebookController
    // copy("test:"+(!notebookController.view.hidden && notebookController.mindmapView.selViewLst.length))
    if (!notebookController.view.hidden && notebookController.mindmapView && notebookController.mindmapView.selViewLst.length) {
      let selViewLst = notebookController.mindmapView.selViewLst
      // let focusNotes = 
      return selViewLst.map(tem=>{
        return tem.note.note
      })
    }else if(this.currentDocController().focusNote){
      return [this.currentDocController().focusNote]
    }else{
      //如果两个上面两个都没有，那就可能是小窗里打开的
      return [notebookController.focusNote]
    }
  }

  static cloneNote(noteid) {
    let targetNote = this.getNoteById(noteid)
    if (!targetNote) {
      this.showHUD("Note not exists!")
      return undefined
    }
    let note = this.data.cloneNotesToTopic([targetNote], targetNote.notebookId)
    return note[0]
  }

  static cloneAndMerge(currentNote,targetNoteId) {
    let clonedNoteId = this.cloneNote(targetNoteId)
    currentNote.merge(this.getNoteById(clonedNoteId))
  }

  static postNotification(name,userInfo) {
    NSNotificationCenter.defaultCenter().postNotificationNameObjectUserInfo(name, this.focusWindow, userInfo)
  }

  static moveElement(arr, element, direction) {
      // 获取元素的索引
      var index = arr.indexOf(element);
      if (index === -1) {
          this.showHUD('Element not found in array');
          return;
      }
      switch (direction) {
          case 'up':
              if (index === 0) {
                  this.showHUD('Element is already at the top');
                  return;
              }
              // 交换元素位置
              [arr[index], arr[index - 1]] = [arr[index - 1], arr[index]];
              break;
          case 'down':
              if (index === arr.length - 1) {
                  this.showHUD('Element is already at the bottom');
                  return;
              }
              // 交换元素位置
              [arr[index], arr[index + 1]] = [arr[index + 1], arr[index]];
              break;
          case 'top':
              // 移除元素
              arr.splice(index, 1);
              // 添加到顶部
              arr.unshift(element);
              break;
          default:
              this.showHUD('Invalid direction');
      }
  }

  static getVarInfo(text) {
    let config = {}
    let hasClipboardText = text.includes("{{clipboardText}}")
    let hasSelectionText = text.includes("{{selectionText}}")
    if (hasClipboardText) {
      config.clipboardText = this.clipboardText()
    }
    if (hasSelectionText) {
      config.selectionText = this.currentDocController().selectionText
    }
    return config
  }

  static replacVar(text,varInfo) {
    let vars = Object.keys(varInfo)
    let original = text
    for (let i = 0; i < vars.length; i++) {
      const variable = vars[i];
      const variableText = varInfo[variable]
      original = original.replace(`{{${variable}}}`,variableText)
    }
    // copy(original)
    return original
  }

  static detectAndReplace(text,element=undefined) {
    let config = this.getVarInfo(text)
    if (element !== undefined) {
      config.element = element
    }
    return this.replacVar(text,config)
  }

  static checkHeight(height){
    if (height > 400) {
      return 400
    }else if(height < 80){
      return 80
    }else{
      return height
    }
  }
  static addObserver(observer,selector,name){
    NSNotificationCenter.defaultCenter().addObserverSelectorName(observer, selector, name);
  }
  static removeObserver(observer,name){
    NSNotificationCenter.defaultCenter().removeObserverName(observer, name);
  }
  static studyMode(){
    return this.studyController().studyMode
  }
  static openURL(url){
    this.app.openURL(NSURL.URLWithString(url));
  }
  static initChatAIController(){
    this.chatController = chatglmController.new();
    MNUtil.studyView.addSubview(this.chatController.view)
  }
  static initNotifyController(){
    this.notifyController = notificationController.new();
    MNUtil.studyView.addSubview(this.notifyController.view)
  }
  static initDynamicController(){
    this.dynamicController = dynamicController.new();
    MNUtil.studyView.addSubview(this.dynamicController.view)
  }
  /**
   * 
   * @param {MbBookNote|MNNote} note 
   */
  static async getTextForSearch(note,OCR_enabled=false) {
    let order = chatAIConfig.config.searchOrder
    let text
    for (let index = 0; index < order.length; index++) {
      const element = order[index];
      switch (element) {
        case 1:
          if (note.noteTitle && note.noteTitle !== "") {
            text = note.noteTitle
          }
          break;
        case 2:
          if (OCR_enabled && note.excerptPic && !note.textFirst) {
            let image = MNUtil.getMediaByHash(note.excerptPic.paint)
            text = await chatAINetwork.getTextOCR(image)
            // text = ""
          }else if (this.noteHasExcerptText(note)) {
            text = note.excerptText
          }
          //如果都不满足，此时text依然为undefined
          break;
        case 3:
          let commentText
          let commentImage = undefined
          //用find主要是保证得到想要的元素之后就会直接返回，而不是继续执行
          //但是find处理不了异步的情况
          let comment = note.comments.find(comment=>{
            switch (comment.type) {
              case "TextNote":
                if (/^marginnote\dapp:\/\//.test(comment.text)) {
                  return false
                }else{
                  commentText = comment.text
                  return true
                }
              case "HtmlNote":
                commentText = comment.text
                return true
              case "LinkNote":
                if (comment.q_hpic && !note.textFirst) {
                  if (OCR_enabled) {
                    commentImage = MNUtil.getMediaByHash(comment.q_hpic.paint)
                    // commentText = await chatAINetwork.getTextOCR(commentImage)
                    return true
                  }else{
                    return false
                  }
                }else{
                  commentText = comment.q_htext
                  return true
                }
              case "PaintNote":
                if (OCR_enabled && comment.paint) {
                  commentImage = MNUtil.getMediaByHash(comment.paint)
                  // commentText = await chatAINetwork.getTextOCR(commentImage)
                  return true
                }
                return false
              default:
                return false
            }
          })
          if (commentImage) {
            MNUtil.showHUD("should OCR")
            commentText = await chatAINetwork.getTextOCR(commentImage)
            // MNUtil.copy(commentText)
          }
          if (commentText && commentText.length) {
            text = commentText
          }
          break;
        default:
          break;
      }
      if (text) {
        return text
      }
    }
  return undefined
  }
  static getToday() {
    // 创建一个新的Date对象，默认情况下它会包含当前日期和时间
    const today = new Date();

    // 获取日
    const day = today.getDate();
    return day
  }
  static switchColor(on123){
    let color = on123 ? "#457bd3" : "#9bb2d6"
    return this.hexColorAlpha(color,0.8)
  }
  static countWords(str) {
    const chinese = Array.from(str)
        .filter(ch => /[\u4e00-\u9fa5]/.test(ch))
        .length
    
    const english = Array.from(str)
        .map(ch => /[a-zA-Z0-9\s]/.test(ch) ? ch : ' ')
        .join('').split(/\s+/).filter(s => s)
        .length

    return chinese + english
  }
  static animate(func,time = 0.2) {
    UIView.animateWithDurationAnimationsCompletion(time,func)
  }

  static animateWithCompletion(func,funcWhenCompleted,time = 0.2) {
    UIView.animateWithDurationAnimationsCompletion(time,func,funcWhenCompleted)
  }
  static checkSender(sender,window){
    return this.app.checkNotifySenderInWindow(sender, window)
  }
  static getFileName(fullPath) {
      // 找到最后一个'/'的位置
      let lastSlashIndex = fullPath.lastIndexOf('/');

      // 从最后一个'/'之后截取字符串，得到文件名
      let fileName = fullPath.substring(lastSlashIndex + 1);

      return fileName;
  }
  static isfileExists(path) {
    return NSFileManager.defaultManager().fileExistsAtPath(path)
  }
  static getCurrentFile() {
    let docFile = this.currentDocController().document
    // copy(docFile.pathFile)
    let fullPath
    if (docFile.fullPathFileName) {
      fullPath = docFile.fullPathFileName
    }else{
      let folder = this.app.documentPath
      fullPath = folder+"/"+docFile.pathFile
      if (docFile.pathFile.startsWith("$$$MNDOCLINK$$$")) {
        let fileName = this.getFileName(docFile.pathFile)
        fullPath = this.app.tempPath + fileName
        // fullPath = docFile.pathFile.replace("$$$MNDOCLINK$$$", "/Users/linlifei/")
      }
    }

    if (!this.isfileExists(fullPath)) {
      MNUtil.showHUD("Invalid file: "+docFile.pathFile)
      return undefined
    }
    // copy(fullPath)
    let fileName = this.getFileName(fullPath)
    return{
      name:fileName,
      path:fullPath,
      md5:docFile.docMd5
    }
  }
  static getNoteFileName(noteId) {
    let note = MNUtil.getNoteById(noteId)
    let docFile = MNUtil.getDocById(note.docMd5)
    if (!docFile) {
      MNUtil.showHUD("No file")
      return undefined
    }
    return MNUtil.getFileName(docFile.pathFile)
  }
  static getNoteFile(noteId) {
    let note = MNUtil.getNoteById(noteId)
    let docFile = MNUtil.getDocById(note.docMd5)
    if (!docFile) {
      MNUtil.showHUD("No file")
      return undefined
    }
    let fullPath
    if (docFile.fullPathFileName) {
      fullPath = docFile.fullPathFileName
    }else{
      let folder = MNUtil.documentFolder
      let fullPath = folder+"/"+docFile.pathFile
      if (docFile.pathFile.startsWith("$$$MNDOCLINK$$$")) {
        let fileName = this.getFileName(docFile.pathFile)
        fullPath = Application.sharedInstance().tempPath + fileName
        // fullPath = docFile.pathFile.replace("$$$MNDOCLINK$$$", "/Users/linlifei/")
      }
    }
    if (!MNUtil.isfileExists(fullPath)) {
      MNUtil.showHUD("Invalid file: "+docFile.pathFile)
      return undefined
    }
    // copy(fullPath)
    let fileName = this.getFileName(fullPath)
    return{
      name:fileName,
      path:fullPath,
      md5:docFile.docMd5
    }
  }
  static isValidJSON(jsonString){
    try{
        var json = JSON.parse(jsonString);
        if (json && typeof json === "object") {
            return true;
        }
    }
    catch(e){
        return false;
    }
    return false;
  }
  static noteHasExcerptText(note) {
    //摘要文本必须存在并且不为空
    //考虑到可能存在图片摘要，因此要么图片摘要不存在，要么图片摘要已被转为文本
    return note.excerptText && note.excerptText !== "" && (!note.excerptPic || note.textFirst)
  }
  static strCode(str) {  //获取字符串的字节数
    var count = 0;  //初始化字节数递加变量并获取字符串参数的字符个数
    var cn = [8211, 8212, 8216, 8217, 8220, 8221, 8230, 12289, 12290, 12296, 12297, 12298, 12299, 12300, 12301, 12302, 12303, 12304, 12305, 12308, 12309, 65281, 65288, 65289, 65292, 65294, 65306, 65307, 65311]
    var half = [32, 33, 34, 35, 36, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 58, 59, 60, 61, 62, 63, 64, 91, 92, 93, 94, 95, 96, 123, 124, 125, 126,105,108,8211]
    if (str) {  //如果存在字符串，则执行
        let len = str.length; 
        for (var i = 0; i < len; i++) {  //遍历字符串，枚举每个字符
          let charCode = str.charCodeAt(i)
            if (charCode>=65 && charCode<=90) {
              count += 1.5;  //大写
            } else if (half.includes(charCode)) {
              count +=0.45
            } else if (cn.includes(charCode)) {
              count +=0.8
            }else if (charCode > 255) {  //字符编码大于255，说明是双字节字符(即是中文)
                count += 2;  //则累加2个
            }else{
                count++;  //否则递加一次
            }
        }
        return count;  //返回字节数
    } else {
        return 0;  //如果参数为空，则返回0个
    }
  }
  static getRandomElement(arr) {
    if (arr.length === 1) {
      return arr[0]
    }
    if (arr && arr.length) {
      const randomIndex = Math.floor(Math.random() * arr.length);
      return arr[randomIndex];
    }
    return ""; // 或者抛出一个错误，如果数组为空或者未定义
  }
  static focusNoteInMindMapById(noteId){
    this.studyController().focusNoteInMindMapById(noteId)
  }
  static async delay (seconds) {
    return new Promise((resolve, reject) => {
      NSTimer.scheduledTimerWithTimeInterval(seconds, false, function () {
        resolve()
      })
    })
  }
  static getFile(path) {
    return NSData.dataWithContentsOfFile(path)
  }
  static clearCurrent(){
    this.currentNoteId = undefined
    this.currentSelection = undefined
  }
  static refreshCurrent(){
    this.currentSelection = this.currentDocController().selectionText
    let focusNote = this.getFocusNote()
    if (focusNote) {
      this.currentNoteId = focusNote.noteId
    }else{
      this.currentNoteId = undefined
    }
  }
  static getPopoverAndPresent(sender,commandTable,width=100,position=2) {
    var menuController = MenuController.new();
    menuController.commandTable = commandTable
    menuController.rowHeight = 35;
    menuController.preferredContentSize = {
      width: width,
      height: menuController.rowHeight * menuController.commandTable.length
    };
    //左 0
    //下 1，3
    //上 2
    //右 4
    var popoverController = new UIPopoverController(menuController);
    var r = sender.convertRectToView(sender.bounds,MNUtil.studyView);
    popoverController.presentPopoverFromRect(r, MNUtil.studyView, position, true);
    return popoverController
  }
  /**
   * 
   * @returns {Boolean}
   */
  static checkSubscribe(count = true){
    // return true
    if (typeof subscriptionConfig !== 'undefined') {
      let res = subscriptionConfig.checkSubscribed(count)
      return res
    }else{
      this.showHUD("Please install 'MN Subscription' first!")
      return false
    }
  }
  /**
   * 
   * @param {MbBookNote} note
   * @returns {String[]}
   */
  static extractTagsFromNote(note) {
    // 用于存储所有标签的数组
    const tags = [];

    // 检查 note 是否有 comments 属性，且其为数组
    if (note.comments && Array.isArray(note.comments)) {
      // 遍历 comments 数组
      note.comments.forEach(comment => {
        // 检查评论的类型是否为 "TextNote"
        if (comment.type === "TextNote" && typeof comment.text === "string") {
          // 使用正则表达式查找所有标签（格式为 #tagname）
          const tagMatches = comment.text.match(/#[\w\u4e00-\u9fff]+/g);
          if (tagMatches) {
            // 将找到的标签添加到 tags 数组中
            tags.push(...tagMatches);
          }
        }
      });
    }

    // 返回去重后的标签数组
    return [...new Set(tags)];
  }
  /**
   * 
   * @param {*} arr1 
   * @param {*} arr2 
   * @returns {String[]}
   */
  static findCommonElements(arr1, arr2) {
    // 使用 filter() 方法和 includes() 方法找出相同的元素
    const commonElements = arr1.filter(element => arr2.includes(element));
    return commonElements;
  }
  static findKeyByTitle(obj, titleValue) {
    for (let key in obj) {
      if (obj.hasOwnProperty(key)) {
        if (obj[key].title === titleValue) {
          return key;
        }
      }
    }
    return undefined; // 没有找到对应的 key
  }
  static getFuncForAI(funcName) {
    let args = {}
    switch (funcName) {
      case "setTitle":
        args.title = this.genArgument("string", "title that should be set for this card, pure text only")
        return this.genStructure(funcName, args,
          "this function is used to set title for a card"
        )
      case "addComment":
        args.comment = this.genArgument("string", "comment that should be added for this card, pure text only")
        return this.genStructure(funcName, args,
          "this function is used to add a comment for a card"
        )
      case "addTag":
        args.tag = this.genArgument("string", "tag that should be added for this card, single word only, no hyphen is allowed")
        return this.genStructure(funcName, args,
          "this function is used to add a tag for a card"
        )
      case "addChildNote":
        args.title = this.genArgument("string", "title for child note, optional")
        args.content = this.genArgument("string", "content for child note, optional")
        return this.genStructure(funcName, args,
          "this function is used to add a child note for a card"
        )
      case "clearExcerpt":
        return this.genStructure(funcName, args,
          "this function is used to clear the card excerpt, do not pass any argument"
        )
      case "setExcerpt":
        args.excerpt = this.genArgument("string", "excerpt text to be set")
        return this.genStructure(funcName, args,
          "this function is used to set the card excerpt text"
        )
      case "copyMarkdownLink":
        args.title = this.genArgument("string", "title for markdown link")
        return this.genStructure(funcName, args,
          "this function can copy the card link in mardown format, just specify the title for this link"
        )
      case "copyCardURL":
        return this.genStructure(funcName, args,
          "this function can directly copy the card url, do not pass any argument"
        )
      case "copyText":
        args.text = this.genArgument("string", "pure text that need to be copied")
        return this.genStructure(funcName, args,
          "this function can copy any pure text, but not card link"
        )
      case "close":
        return this.genStructure(funcName, args,
          "this function is used to end the coversation, do not pass any argument"
        )
      default:
        break;
    }
  }
  static getFunByIndex(indices,isNote = true) {
    let func = ["setTitle","addComment","copyMarkdownLink","copyCardURL","copyText","close","addTag","addChildNote","clearExcerpt","setExcerpt"]
    let disableForSelection = [true,true,true,true,false,false,true,true,true,true]
    if (isNote) {
      let funcStructures = indices.map(ind=>this.getFuncForAI(func[ind]))
      return funcStructures
    }else{
      let funcStructures = indices.filter(ind=>!disableForSelection[ind]).map(ind=>this.getFuncForAI(func[ind]))
      return funcStructures
    }
  }

  static genArgument(type,description) {
    return{ "type": type, "description": description }
  }
  static genStructure(name,args,description) {
    let funcStructure = {
      "name":name,
      "description":description,
      "parameters": {
        "type": "object",
        "properties": args,
        "required": Object.keys(args),
      }
    }
    return {"type":"function","function":funcStructure}
  }
  /**
   * 
   * @param {Array} resList 
   */
  static checkToolCalls(resList) {
    // MNUtil.copyJSON(resList)
    try {
      
    return resList.some(res=>{
      if (res && res.tool_calls) {
        return true
      }
      return false
    })
    } catch (error) {
      MNUtil.showHUD("Error in checkToolCalls: "+error)
      return false
    }
  }
  /**
   * 
   * @param {String} rawText 
   * @returns {{event:String,data:String}[]}
   */
  static parseEvents(rawText) {
    const lines = rawText.split('\n\n');
    // return lines
    const events = lines.map(line => {
      // 假设每行文本都是以 "event: " 开头，后面跟着事件类型，然后是数据
      const parts = line.split('\n');
      // const eventType = parts[0].split(': ')[1].trim();
      // const eventData = JSON.parse(parts.slice(1).join('\n'));
      let event = parts[0].split(':')[1]
      if (event) {
        let data = parts[1].slice(5)
        return {
          event: event.trim(),
          data: JSON.parse(data)
        };
      }else{
        return {
          event: ""
        };
      }
    });
    return events;
  }
  static getResList(originalText) {
    try {
      
    let res = originalText.split("data:")
    // copy(originalText)
    res.shift()
    if (res.length > 1) {
      res.pop()
    }
    if (JSON.parse(res[0]).choices.length === 0) {
      res.shift()
    }
    // MNUtil.copyJSON(res)
    let resList = res.map(r=>MNUtil.getValidJSON(r).choices[0].delta)
    return resList
    } catch (error) {
      // MNUtil.showHUD("Error in getResList: "+)
      // MNUtil.copy(error)
      return []
    }
  }
  static html(string) {
    return `<!doctype html>
  <html>
  <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0,minimum-scale=1.0, maximum-scale=1.0, user-scalable=no">
    <title>Marked.js</title>
    <style>
  p {
    max-width: 385px;
    margin-top: 0;
  }
  pre {
    background-color: #343541;
    color: white;
    border-radius: 5px;
    max-width: 400px;
    overflow-x: scroll;
    margin-top: 0;
    margin-bottom: 5px;
  }
  img {
    max-width: 385px;
  }
.code-header {
  display: flex;
  justify-content: space-between;
  align-items: center;
  padding: 5px;
}
.code-header button {
  margin-left: auto; /* 这将推动按钮到容器的右边 */
}
pre code.hljs {
  display: block;
  overflow-x: auto;
  padding: 1em
}
code.hljs {
  padding: 3px 5px
}
/*!
  Theme: GitHub Dark
  Description: Dark theme as seen on github.com
  Author: github.com
  Maintainer: @Hirse
  Updated: 2021-05-15

  Outdated base version: https://github.com/primer/github-syntax-dark
  Current colors taken from GitHub's CSS
*/
.hljs {
  color: #c9d1d9;
  background: #000000
}
.hljs-doctag,
.hljs-keyword,
.hljs-meta .hljs-keyword,
.hljs-template-tag,
.hljs-template-variable,
.hljs-type,
.hljs-variable.language_ {
  /* prettylights-syntax-keyword */
  color: #ff7b72
}
.hljs-title,
.hljs-title.class_,
.hljs-title.class_.inherited__,
.hljs-title.function_ {
  /* prettylights-syntax-entity */
  color: #d2a8ff
}
.hljs-attr,
.hljs-attribute,
.hljs-literal,
.hljs-meta,
.hljs-number,
.hljs-operator,
.hljs-variable,
.hljs-selector-attr,
.hljs-selector-class,
.hljs-selector-id {
  /* prettylights-syntax-constant */
  color: #79c0ff
}
.hljs-regexp,
.hljs-string,
.hljs-meta .hljs-string {
  /* prettylights-syntax-string */
  color: #a5d6ff
}
.hljs-built_in,
.hljs-symbol {
  /* prettylights-syntax-variable */
  color: #ffa657
}
.hljs-comment,
.hljs-code,
.hljs-formula {
  /* prettylights-syntax-comment */
  color: #8b949e
}
.hljs-name,
.hljs-quote,
.hljs-selector-tag,
.hljs-selector-pseudo {
  /* prettylights-syntax-entity-tag */
  color: #7ee787
}
.hljs-subst {
  /* prettylights-syntax-storage-modifier-import */
  color: #c9d1d9
}
.hljs-section {
  /* prettylights-syntax-markup-heading */
  color: #1f6feb;
  font-weight: bold
}
.hljs-bullet {
  /* prettylights-syntax-markup-list */
  color: #f2cc60
}
.hljs-emphasis {
  /* prettylights-syntax-markup-italic */
  color: #c9d1d9;
  font-style: italic
}
.hljs-strong {
  /* prettylights-syntax-markup-bold */
  color: #c9d1d9;
  font-weight: bold
}
.hljs-addition {
  /* prettylights-syntax-markup-inserted */
  color: #aff5b4;
  background-color: #033a16
}
.hljs-deletion {
  /* prettylights-syntax-markup-deleted */
  color: #ffdcd7;
  background-color: #67060c
}
.hljs-char.escape_,
.hljs-link,
.hljs-params,
.hljs-property,
.hljs-punctuation,
.hljs-tag {
  /* purposely ignored */
}
  </style>
  </head>
  <body contenteditable="true">${string}<script>
  // window.scrollTo(0,document.body.scrollHeight);

      MathJax = {
          tex: {
              inlineMath: [ ['$','$'], ["\\(","\\)"] ]
          }
      };
      var originalBodyContent = ""
      var isBodyEdited = false;
      document.addEventListener('input', function () {
        isBodyEdited = true;
      });
      function getSelectedTextOrPageText() {
          var selectedText = window.getSelection().toString();
          if (selectedText && selectedText.trim() !== '') {
              return selectedText;
          } else if (isBodyEdited){
              var body = document.body;
              var pageText = body.innerText || body.textContent;
              return pageText.trim();
          } else {
            return ""
          }
      }
      function copyToClipboard(uuid) {
        // 获取代码块的文本内容
        var code = document.getElementById(uuid).innerText;
        // 创建临时的textarea来选择文本
        var dummy = document.createElement('textarea');
        // 将代码块的文本内容赋值到textarea
        dummy.value = code;
        document.body.appendChild(dummy);
        // 选择textarea中的内容
        dummy.select();
        // 执行复制命令
        document.execCommand('copy');
        // 移除临时创建的textarea
        document.body.removeChild(dummy);
      }
  </script>
  <script id="MathJax-script" async src="https://cdn.jsdelivr.net/npm/mathjax@3/es5/tex-chtml.js"></script>
  </body>
  </html>`
  }
  static sum(array) {
    let sum = 0
    for (let i = 0; i < array.length; i++) {
      sum = sum+array[i]
    }
    return sum
  }
  static getNewLoc(gesture){
    let locationToMN = gesture.locationInView(MNUtil.studyView)
    if (!gesture.moveDate) {
      gesture.moveDate = 0
    }
    if ((Date.now() - gesture.moveDate) > 100) {
      let translation = gesture.translationInView(MNUtil.studyView)
      let locationToBrowser = gesture.locationInView(gesture.view.superview)
      let locationToButton = gesture.locationInView(gesture.view)
      let buttonFrame = gesture.view.frame
      let newY = locationToButton.y-translation.y 
      let newX = locationToButton.x-translation.x
      // if (gesture.state !== 3 && Math.abs(translation.y)<20 && Math.abs(translation.x)<20) {
      if (gesture.state !== 3 && (newY<buttonFrame.height+5 && newY>-5 && newX<buttonFrame.width+5 && newX>-5 && Math.abs(translation.y)<20 && Math.abs(translation.x)<20)) {
        gesture.locationToBrowser = {x:locationToBrowser.x-translation.x,y:locationToBrowser.y-translation.y}
        // MNUtil.showHUD(JSON.stringify(gesture.locationToBrowser))
      }
    }
    // MNUtil.showHUD(JSON.stringify(locationToMN))
    if (locationToMN.x <= 0) {
      locationToMN.x = 0
    }
    if (locationToMN.x > MNUtil.studyView.frame.width) {
      locationToMN.x = MNUtil.studyView.frame.width
    }
    gesture.moveDate = Date.now()
    // let location = {x:locationToMN.x - self.locationToButton.x-gesture.view.frame.x,y:locationToMN.y -self.locationToButton.y-gesture.view.frame.y}
    let location = {x:locationToMN.x - gesture.locationToBrowser.x,y:locationToMN.y -gesture.locationToBrowser.y}
    if (location.y <= 0) {
      location.y = 0
    }
    if (location.y>=MNUtil.studyView.frame.height-15) {
      location.y = MNUtil.studyView.frame.height-15
    }
    return location
  }
  /**
   * 
   * @param {string} fullPath 
   * @returns {string}
   */
  static getExtensionFolder(fullPath) {
      // 找到最后一个'/'的位置
      let lastSlashIndex = fullPath.lastIndexOf('/');
      // 从最后一个'/'之后截取字符串，得到文件名
      let fileName = fullPath.substring(0,lastSlashIndex);
      return fileName;
  }
  static checkMNUtilsFolder(fullPath){
    let extensionFolder = this.getExtensionFolder(fullPath)
    return NSFileManager.defaultManager().fileExistsAtPath(extensionFolder+"/marginnote.extension.mnutils/main.js")
  }
  static checkCouldAsk(){
      if (this.notifyController.view.hidden) {
        //如果界面都隐藏了，那肯定可以启动
        this.notifyController.onChat = false
        return true
      }else{
        //如果请求还没结束，或者在聊天模式，则不允许发起对话
        if (this.notifyController.connection || this.notifyController.onChat) {
          MNUtil.showHUD("on output or on chat mode")
          return false
        }else{
          return true
        }
      }
  }
  /**
   * 
   * @param {UIView} view 
   */
  static ensureView(view){
    if (!MNUtil.isDescendantOfStudyView(view)) {
      view.hidden = true
      MNUtil.studyView.addSubview(view)
    }
  }
}

class chatAIConfig {
  // 构造器方法，用于初始化新创建的对象
  constructor(name) {
    this.name = name;
  }
  // static defaultAction
  static isFirst = true
  static his = []
  static currentPrompt
  /** @type {Number[]} */
  static currentFunc
  /** @type {String} */
  static currentTitle
  /** @type {String} */
  static mainPath
  static action = []
  static defaultPrompts = {
      Translate:                      {title: "翻译",context:"请翻译下面这段话：{{context}}"},
      Keyword:                        {title: "关键词",context:"请为下面这段话提取关键词：{{context}}"},
      Title:                          {title: "标题",context:"请为下面这段话生成标题：{{context}}"}
  }
  static defaultConfig = {
    searchOrder: [2,1,3],
    autoAction: false,
    onSelection: true,
    onNote: true,
    onNewExcerpt: true,
    delay: 0,
    ignoreShortText: false,
    notifyLoc: 0,
    toolbar: true,
    model: 0,
    chatglmModel:0,
    promptNames: Object.keys(this.defaultPrompts),
    currentPrompt: Object.keys(this.defaultPrompts)[0],
    apikey: "",
    modelOnReAsk: -1,
    source: 'Built-in',
    url: "https://api.openai.com",
    openaiKey: "",
    tunnel:0,
    moonshotKey : "",
    moonshotModel : 0,
    customKey : "",
    customModel : "",
    customUrl : "",
    customModelIndex : 0,
    claudeKey : "",
    claudeUrl : "https://api.anthropic.com",
    claudeModel : 0,
    geminiKey : "",
    geminiUrl : 'https://generativelanguage.googleapis.com',
    geminiModel : 0,
    dynamic:false,
    dynamicFunc : [],
    dynamicModel : "Default",
    colorConfig: [1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1],
    syncNoteId: "",
    simpleTexKey:"",
    autoExport:false,
    autoImport:false,
    lastSyncTime:0,
    modifiedTime:0
  }
  static defaultDynamicPrompt = {
    "note":"list below is the structure of a card:\n{{card}}",
    "text":"{{context}}"
  }
  static default_usage = {
    usage: 0,
    limit: 100,
    day: 0
  }
  static init(){
    this.prompts = this.getByDefault('MNChatglm_prompts', this.defaultPrompts)
    // MNUtil.copyJSON(this.prompts)
    this.knowledge = this.getByDefault('MNChatglm_knowledge',"")
    this.keys = this.getByDefault('MNChatglm_builtInKeys', {})
    this.config = this.getByDefault('MNChatglm_config',this.defaultConfig)
    this.fileId = this.getByDefault("MNChatglm_fileId", {})
    this.default_usage = {limit:100,day:chatAIUtils.getToday(),usage:0}
    this.usage = this.getByDefault('MNChatglm_usage',this.default_usage)
    this.dynamicPrompt = this.getByDefault("MNChatglm_dynamicPrompt", this.defaultDynamicPrompt)
    this.currentPrompt = this.getConfig("currentPrompt")
    // MNUtil.copyJSON({prompts:this.prompts,config:this.config})
    let currentPrompt = this.prompts[this.currentPrompt]
    if (!currentPrompt) {
      let promptNames = Object.keys(this.prompts)
      this.currentPrompt = promptNames[0]
    }
    this.currentFunc = this.prompts[this.currentPrompt].func ?? []
    this.currentModel = this.prompts[this.currentPrompt].model ?? "Default"
    this.currentTitle = this.prompts[this.currentPrompt].title
    this.defaultModel = this.getDefaultModel()
    this.highlightColor = UIColor.blendedColor(
      UIColor.colorWithHexString("#2c4d81").colorWithAlphaComponent(0.8),
      chatAIUtils.app.defaultTextColor,
      0.8
    );
  }
  static getConfig(key){
    if (this.config[key] !== undefined) {
      return this.config[key]
    }else{
      return this.defaultConfig[key]
    }
  }
  // static currentFunc(){
  //   let currentFunc = this.prompts[this.currentPrompt].func ?? []
  //   return currentFunc
  // }
  /**
   * 
   * @param {MbBookNote} note
   */
  static expandesConfig(note,config,orderedKeys=undefined,exclude=undefined) {
    let keys
    if (orderedKeys) {
      keys = orderedKeys
    }else{
      keys = Object.keys(config)
    }
    keys.forEach((key)=>{
      let subConfig = config[key]
      if (typeof subConfig === "object") {
        let child = chatAIUtils.createChildNote(note,key)
        this.expandesConfig(child, subConfig,undefined,exclude)
      }else{
        if (exclude) {
          if (key !== exclude) {
            chatAIUtils.createChildNote(note,key,config[key])
          }
        }else{
          chatAIUtils.createChildNote(note,key,config[key])
        }
      }
    })
  }
  static getDefaultActionKeys() {
    let actions = this.getActions()
    return Object.keys(actions)
  }
  static save(key,ignoreExport = false,synchronize = true) {
      switch (key) {
        case "MNChatglm_builtInKeys":
          NSUserDefaults.standardUserDefaults().setObjectForKey(this.keys,'MNChatglm_builtInKeys')
          break;
        case "MNChatglm_config":
          NSUserDefaults.standardUserDefaults().setObjectForKey(this.config,"MNChatglm_config")
          this.config.modefiedTime = Date.now()
          if (!ignoreExport && this.getConfig("autoExport")) {
            this.export(false)
          }
          break;
        case "MNChatglm_prompts":
          NSUserDefaults.standardUserDefaults().setObjectForKey(this.prompts,"MNChatglm_prompts")
          this.config.modefiedTime = Date.now()
          if (!ignoreExport && this.getConfig("autoExport")) {
            this.export(false)
          }
          break;
        case "MNChatglm_fileId":
          NSUserDefaults.standardUserDefaults().setObjectForKey(this.fileId,"MNChatglm_fileId")
          break;
        case "MNChatglm_knowledge":
          NSUserDefaults.standardUserDefaults().setObjectForKey(this.knowledge,"MNChatglm_knowledge")
          this.config.modefiedTime = Date.now()
          if (!ignoreExport && this.getConfig("autoExport")) {
            this.export(false)
          }
          break;
        case "MNChatglm_dynamicPrompt":
          NSUserDefaults.standardUserDefaults().setObjectForKey(this.dynamicPrompt,"MNChatglm_dynamicPrompt")
          this.config.modefiedTime = Date.now()
          if (!ignoreExport && this.getConfig("autoExport")) {
            this.export(false)
          }
          break;
        default:
          MNUtil.showHUD("Not supported")
          break;
      }
    if (synchronize) {
      NSUserDefaults.standardUserDefaults().synchronize()
    }
  }
  static async export(alert = true){
    // chatAIConfig.his.push("export")
    // MNUtil.showHUD("his: "+chatAIConfig.his.length)
    // MNUtil.copyJSON(chatAIConfig.his)
    // MNUtil.showHUD("export")
    if (!chatAIUtils.checkSubscribe(true)) {
      return false
    }
    let noteId = this.config.syncNoteId
    let lastSyncTime = this.config.lastSyncTime ?? 0
    let modifiedTime = this.config.modifiedTime ?? 0
    let latestTime = Math.max(lastSyncTime,modifiedTime)
    let focusNote = MNUtil.getNoteById(noteId)
    if (!focusNote) {
      focusNote = MNNote.getFocusNote()
    }
    let modifiedDate = Date.parse(focusNote.modifiedDate ?? focusNote.createDate)
    let confirm = false
    // MNUtil.copy(modifiedDate)
    // MNUtil.copyJSON(focusNote.modifiedDate)
    if (latestTime > modifiedDate) {
      // MNUtil.showHUD("should sync")
      confirm = true
    }else{
      // MNUtil.showHUD("should comfirm")
      if (alert) {
        confirm = await MNUtil.confirm("Newer config from note!\n卡片配置较新！","Overwrite?\n是否覆盖？")
      }
    }
    if (!confirm) {
      return false
    }
    this.config.lastSyncTime = Date.now()+5
    this.config.syncNoteId = focusNote.noteId
    let config = {config:this.config,prompts:this.prompts,knowledge:this.knowledge,dynamicPrompt:this.dynamicPrompt}
    MNUtil.undoGrouping(()=>{
      focusNote.excerptText = JSON.stringify(config,null,2)
      focusNote.noteTitle = "MN ChatAI Config"
    })
    return true
    // MNUtil.copyJSON(config)
  }
  static async import(alert = true){
    // chatAIConfig.his.push("import")
    // MNUtil.showHUD("his: "+chatAIConfig.his.length)
    // MNUtil.copyJSON(chatAIConfig.his)

    if (!chatAIUtils.checkSubscribe(true)) {
      return false
    }
    let noteId = this.config.syncNoteId
    let lastSyncTime = this.config.lastSyncTime ?? 0
    let modifiedTime = this.config.modifiedTime ?? 0
    let latestTime = Math.max(lastSyncTime,modifiedTime)
    let focusNote = MNUtil.getNoteById(noteId)
    if (!focusNote) {
      focusNote = chatAIUtils.getFocusNote()
    }
    if (focusNote.noteTitle !== "MN ChatAI Config") {
      MNUtil.showHUD("Invalid note title!")
      return false
    }
    let modifiedDate = Date.parse(focusNote.modifiedDate)
    let confirm = false
    // MNUtil.copy(modifiedDate)
    // MNUtil.copyJSON(focusNote.modifiedDate)
    if (latestTime < modifiedDate) {
      // MNUtil.showHUD("should sync")
      confirm = true
    }else{
      // MNUtil.showHUD("should comfirm")
      if (alert) {
        confirm = await MNUtil.confirm("Older config from note!\n卡片配置较旧！","Overwrite?\n是否覆盖？")
      }
    }
    if (!confirm) {
      return false
    }
    let config = JSON.parse(focusNote.excerptText)
    if ("config" in config && "prompts" in config && "knowledge" in config && "dynamicPrompt" in config) {
      this.config = config.config
      this.config.lastSyncTime = Date.now()
      // this.config.modifiedTime = modifiedDate
      // let prompts = {}
      // this.config.promptNames.map(promptName=>{
      //   prompts[promptName] = config[promptName]
      // })
      this.prompts = config.prompts
      // MNUtil.copyJSON(prompts)
      this.knowledge = config.knowledge
      this.dynamicPrompt = config.dynamicPrompt
      this.currentPrompt = this.config.currentPrompt
      chatAIUtils.chatController.lastLength = this.config.promptNames.length
      chatAIUtils.chatController.refreshView("autoActionView")
      chatAIUtils.chatController.refreshView("syncView")
      chatAIUtils.chatController.refreshView("configView")
      chatAIUtils.chatController.refreshView("modelView")
      chatAIUtils.chatController.refreshView("advanceView")
      this.save("MNChatglm_dynamicPrompt",true)
      this.save("MNChatglm_knowledge",true)
      this.save("MNChatglm_prompts",true)
      this.save("MNChatglm_config",true)
      return true
    }else{
      MNUtil.showHUD("Invalid config in note!")
      return false
    }
  }
 static async sync(){
 try {
  

    if (!chatAIUtils.checkSubscribe(true)) {
      return false
    }
    let noteId = this.config.syncNoteId
    let lastSyncTime = this.config.lastSyncTime ?? 0
    let modifiedTime = this.config.modifiedTime ?? 0
    let latestTime = Math.max(lastSyncTime,modifiedTime)
    let focusNote = MNUtil.getNoteById(noteId)
    if (!focusNote) {
      focusNote = chatAIUtils.getFocusNote()
    }
    let isVaildNoteConfig = false
    let modifiedDate = Date.parse(focusNote.modifiedDate ?? focusNote.createDate)
    if (focusNote.noteTitle === "MN ChatAI Config") {
      let noteConfig = JSON.parse(focusNote.excerptText)
      isVaildNoteConfig = ("config" in noteConfig && "prompts" in noteConfig && "knowledge" in noteConfig && "dynamicPrompt" in noteConfig)
      
    }
    if (latestTime < modifiedDate && isVaildNoteConfig) {
      // MNUtil.showHUD("should import")
      let noteConfig = JSON.parse(focusNote.excerptText)
      this.config = noteConfig.config
      this.config.lastSyncTime = Date.now()
      this.prompts = noteConfig.prompts
      this.knowledge = noteConfig.knowledge
      this.dynamicPrompt = noteConfig.dynamicPrompt
      chatAIUtils.chatController.refreshView("autoActionView")
      chatAIUtils.chatController.refreshView("syncView")
      chatAIUtils.chatController.refreshView("configView")
      chatAIUtils.chatController.refreshView("modelView")
      chatAIUtils.chatController.refreshView("advanceView")
      this.save("MNChatglm_dynamicPrompt",true)
      this.save("MNChatglm_knowledge",true)
      this.save("MNChatglm_prompts",true)
      this.save("MNChatglm_config",true)
      MNUtil.showHUD("Sync Success (import)!")
      return true
    }else{
      // MNUtil.showHUD("should export")
      this.config.lastSyncTime = Date.now()+5
      this.config.syncNoteId = focusNote.noteId
      let config = {config:this.config,prompts:this.prompts,knowledge:this.knowledge,dynamicPrompt:this.dynamicPrompt}
      MNUtil.undoGrouping(()=>{
        focusNote.excerptText = JSON.stringify(config,null,2)
        focusNote.noteTitle = "MN ChatAI Config"
      })
      chatAIUtils.chatController.refreshView("syncView")
      MNUtil.showHUD("Sync Success (export)!")
      return true
    }
 } catch (error) {
  MNUtil.showHUD(error)
 }
 }
  static get(key) {
    return NSUserDefaults.standardUserDefaults().objectForKey(key)
  }

  static getByDefault(key,defaultValue) {
    let value = NSUserDefaults.standardUserDefaults().objectForKey(key)
    if (value === undefined) {
      NSUserDefaults.standardUserDefaults().setObjectForKey(defaultValue,key)
      return defaultValue
    }
    return value
  }

  static remove(key) {
    NSUserDefaults.standardUserDefaults().removeObjectForKey(key)
  }
  static reset(){
    this.prompts = this.defaultPrompts
    this.config.promptNames = Object.keys(this.prompts)
    this.currentPrompt = this.config.promptNames[0]
    this.dynamicPrompt = this.defaultDynamicPrompt

    this.save('MNChatglm_config')
    this.save('MNChatglm_prompts')
    this.remove("MNChatglm_knowledge")
    this.remove("MNChatglm_dynamicPrompt")
  }
  static setCurrentPrompt(prompt){
    this.currentPrompt = prompt//是promptKey
    this.config.currentPrompt = prompt
    this.currentFunc = this.prompts[this.currentPrompt].func ?? []
    this.currentModel = this.prompts[this.currentPrompt].model ?? "Default"
    this.currentTitle = this.prompts[this.currentPrompt].title
    NSUserDefaults.standardUserDefaults().setObjectForKey(this.config,"MNChatglm_config")
  }
  static getUsage() {
    let today = chatAIUtils.getToday()
    this.usage.limit = 100
    if (today !== this.usage.day) {
      this.usage.usage = 0
      this.usage.day = today
    }
    return this.usage
  }
  static modelNames(source){
    switch (source) {
      case "ChatGLM":
        return ["glm-3-turbo","glm-4","glm-4v"]
      case "Gemini":
        return ["gemini-pro","gemini-pro-vision","gemini-1.5-pro-latest","gemini-1.5-flash-latest"]
      case "ChatGPT":
        let models = ["gpt-3.5-turbo","gpt-3.5-turbo-1106","gpt-3.5-turbo-0125","gpt-3.5-turbo-16k","gpt-4","gpt-4o","gpt-4-32k","gpt-4-turbo","gpt-4-turbo-preview","gpt-4-1106-preview","gpt-4-0125-preview","gpt-4-vision-preview"]
        let additionalModels = this.config.customModel.split(",").map(model=>model.trim()).filter(model=>!models.includes(model))
        return models.concat(additionalModels)
      case "KimiChat":
        return ["moonshot-v1-8k","moonshot-v1-32k","moonshot-v1-128k"]
      case "Claude":
        return ["claude-3-haiku-20240307","claude-3-sonnet-20240229","claude-3-opus-20240229"]
      case "Custom":
        return this.config.customModel.split(",").map(model=>model.trim())
      default:
        break;
    }
  }
  static saveApiKey(apikey,url){
    MNUtil.showHUD("Save APIKey for "+this.config.source)
    switch (this.config.source) {
      case "ChatGLM":
        this.config.apikey = apikey
        break;
      case "Claude":
        this.config.claudeKey = apikey
        break;
      case "Gemini":
        this.config.geminiKey = apikey
        break;
      case "ChatGPT":
        this.config.openaiKey = apikey
        if (url) {
          this.config.url = url
        }
        break;
      case "KimiChat":
        this.config.moonshotKey = apikey
        break;
      case "Custom":
        this.config.customKey = apikey
        if (url) {
          this.config.customUrl = url
        }
        break
      default:
        break;
    }
    NSUserDefaults.standardUserDefaults().setObjectForKey(this.config,"MNChatglm_config")
  }
  static getDefaultModel(){
    let models = this.modelNames(this.config.source)
    switch (this.config.source) {
      case "ChatGLM":
        return models[this.getConfig("chatglmModel")]
      case "Claude":
        return models[this.getConfig("claudeModel")]
      case "Gemini":
        return models[this.getConfig("geminiModel")]
      case "ChatGPT":
        return models[this.getConfig("model")]
      case "KimiChat":
        return models[this.getConfig("moonshotModel")]
      case "Custom":
        return models[this.getConfig("customModelIndex")]
      default:
        break;
    }
  }
  static getApiKey(source = this.config.source){
    switch (source) {
      case "ChatGLM":
        return this.getConfig("apikey")
      case "Claude":
        return this.getConfig("claudeKey")
      case "Gemini":
        return this.getConfig("geminiKey")
      case "ChatGPT":
        return this.getConfig("openaiKey")
      case "KimiChat":
        return this.getConfig("moonshotKey")
      case "Custom":
        return this.getConfig("customKey")
      case "Built-in":
        return ""
      default:
        break;
    }
  }
  static getConfigFromSource(source = this.config.source){
    let models = this.modelNames(source)
    let config = {source:source}
    switch (source) {
      case "ChatGLM":
        config.key = this.getConfig("apikey")
        config.url = "https://open.bigmodel.cn/api/paas/v4/chat/completions"
        config.model = models[this.getConfig("chatglmModel")]
        return config
      case "Claude":
        config.key = this.getConfig("claudeKey"),
        config.url = this.getConfig("claudeUrl")+"/v1/messages",
        config.model = models[this.getConfig("claudeModel")]
        return config
      case "Gemini":
        config.key = this.getConfig("geminiKey"),
        config.url = this.getConfig("geminiUrl"),
        config.model = models[this.getConfig("geminiModel")]
        return config
      case "ChatGPT":
        config.key = this.getConfig("openaiKey"),
        config.url = this.getConfig("url")+`/v1/chat/completions`,
        config.model = models[this.getConfig("model")]
        return config
      case "KimiChat":
        config.key = this.getConfig("moonshotKey"),
        config.url = "https://api.moonshot.cn/v1/chat/completions",
        config.model = models[this.getConfig("moonshotModel")]
        return config
      case "Custom":
        config.key = this.getConfig("customKey"),
        config.url = this.getConfig("customUrl"),
        config.model = models[this.getConfig("customModelIndex")]
        return config
      case "Built-in":
        config.key = ""
        return config
      default:
        break;
    }
  }
  static getConfigFromPrompt(prompt = this.currentPrompt){
    let promptConfig = this.prompts[prompt]
    let promptModel = promptConfig.model
    let config
    if (promptModel) {
      let modelConfig = promptModel.split(":")
      if (modelConfig[0] !== "Default") {
        let source = modelConfig[0]
        config = this.getConfigFromSource(source)
        if (modelConfig.length === 2) {
          config.model = modelConfig[1]
        }
      }else{
        config = this.getConfigFromSource()
      }
    }else{
      config = this.getConfigFromSource()
    }
    config.title = promptConfig.title
    if (promptConfig.func) {
      config.func = promptConfig.func
    }
    return config
  }
  static getDynmaicConfig(){
    // let promptConfig = this.dynamicPrompt
    let promptModel = this.getConfig("dynamicModel")
    let dynamicFunc = this.getConfig("dynamicFunc")
    let config
    let modelConfig = promptModel.split(":")
    if (modelConfig[0] !== "Default") {
      let source = modelConfig[0]
      config = this.getConfigFromSource(source)
      if (modelConfig.length === 2) {
        config.model = modelConfig[1]
      }
    }else{
      config = this.getConfigFromSource()
    }
    config.title = "Dynamic"
    if (dynamicFunc.length) {
      config.func = dynamicFunc
    }
    return config
  }
  /**
   * 
   * @param {NSData} fileData 
   * @param {UIView} filename 
   */
  static exportFile(fileData,filename) {
    fileData.writeToFileAtomically(chatAIUtils.app.tempPath+"/"+filename, false)
    let UTI = "public.sqlite"
    chatAIUtils.app.saveFileWithUti(chatAIUtils.app.tempPath+"/"+filename, UTI)
    let tem = NSData.dataWithContentsOfFile(chatAIUtils.app.dbPath+"/vector.sqlite")
    tem.writeToFileAtomically(chatAIUtils.app.dbPath+"/export.sqlite", false)
  }
  static getOrderText() {
    let order = this.getConfig("searchOrder")
    if (order[0] == 4) {
      return 'Order: (Title) + (Excerpt → Comment)'
    }
    let orderNumber = `${order[0]}${order[1]}${order[2]}`
    switch (orderNumber) {
      case "123":
        return 'Order: Title → Excerpt → Comment'
      case "132":
        return 'Order: Title → Comment → Excerpt'
      case "213":
        return 'Order: Excerpt → Title → Comment'
      case "231":
        return 'Order: Excerpt → Comment → Title'
      case "312":
        return 'Order: Comment → Title → Excerpt'
      case "321":
        return 'Order: Comment → Excerpt → Title'
      default:
        return "123";
    }
  }
}

class chatAINetwork {
  constructor(name) {
    this.name = name;
  }
  static requestWithURL(url){
    return NSMutableURLRequest.requestWithURL(NSURL.URLWithString(url))
  }
  static genNSURL(url) {
    return NSURL.URLWithString(url)
  }
  static async sendRequest(request){
    const queue = NSOperationQueue.mainQueue()
    return new Promise((resolve, reject) => {
      NSURLConnection.sendAsynchronousRequestQueueCompletionHandler(
        request,
        queue,
        (res, data, err) => {
          if (err.localizedDescription){
            MNUtil.showHUD(err.localizedDescription)
            reject()
          }
          // if (res.statusCode() === 200) {
          //   MNUtil.showHUD("OCR success")
          // }else{
          //   MNUtil.showHUD("Error in OCR")
          // }
          const result = NSJSONSerialization.JSONObjectWithDataOptions(
            data,
            1<<0
          )
          if (NSJSONSerialization.isValidJSONObject(result)){
            resolve(result)
          }
          resolve(result)
        }
      )
  })
  }
  static async fetch (url,options = {}){
    const request = this.initRequest(url, options)
    const res = await this.sendRequest(request)
    return res
  }
/**
 * 
 * @param {*} url 
 * @param {*} options 
 * @param {NSData} imageData 
 * @returns 
 */
static initOCRRequest (url,options,imageData) {
  const request = NSMutableURLRequest.requestWithURL(this.genNSURL(url))
  // try {
  request.setHTTPMethod("Post")
  request.httpShouldHandleCookies = false
  // request.setCachePolicy(4)
  // request.setTimeoutInterval(options.timeout ?? 10)
  let boundary = "------------------------PN7UsJiL7Z78DgpbkjKEWE"//NSUUID.UUID().UUIDString()

  const headers = {
    "User-Agent": "curl/8.4.0",
    "Accept-Encoding":"*",
    "Accept-Language":"*",
    "Connection":"close",
    "Content-Type": "multipart/form-data; boundary="+boundary,
    Accept: "*/*"
  }
  request.setAllHTTPHeaderFields({
    ...headers,
    ...(options.headers ?? {})
  })

  
  let body = NSMutableData.new()
  let filePart = NSData.dataWithStringEncoding(`--${boundary}\r\nContent-Disposition: form-data; name="file"; filename="Snipaste_2024-02-18_00-32-59.png"\r\nContent-Type: image/png\r\n\r\n`, 4)
  body.appendData(filePart)
  body.appendData(imageData)
  
  let endBoundary = NSData.dataWithStringEncoding(`\r\n--${boundary}--\r\n`, 4)
  body.appendData(endBoundary)
  // copy(body.base64Encoding())
  request.setHTTPBody(body)
  return request
}

static initFileRequest (url,options,purpose,fileName,fileData) {
  const request = NSMutableURLRequest.requestWithURL(this.genNSURL(url))
  try {
  request.setHTTPMethod("Post")
  request.setTimeoutInterval(options.timeout ?? 10)
  let boundary = NSUUID.UUID().UUIDString()

  const headers = {
    "User-Agent":
      "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.1 Safari/605.1.15",
    "Content-Type": "multipart/form-data; boundary="+boundary,
    Accept: "application/json"
  }
  request.setAllHTTPHeaderFields({
    ...headers,
    ...(options.headers ?? {})
  })
  let body = NSMutableData.new()
  let purposeData = NSData.dataWithStringEncoding(purpose, "utf8")
  let purposePart = NSData.dataWithStringEncoding(`
--${boundary}
Content-Disposition: form-data; name="purpose"

`, "utf8")
  let filePart = NSData.dataWithStringEncoding(`
--${boundary}
Content-Disposition: form-data; name="file"; filename="${fileName}"
Content-Type: application/pdf

`, "utf8")
  body.appendData(purposePart)
  body.appendData(purposeData)
  body.appendData(filePart)
  body.appendData(fileData)
  let endBoundary = NSData.dataWithStringEncoding(`--${boundary}--`, "utf8")
  body.appendData(endBoundary)
  request.setHTTPBody(body)
  return request
    } catch (error) {
    MNUtil.showHUD("Error in initFileRequest: "+error)
    return request
  }
}
  /**
   * @param {NSData} image 
   * @returns 
   */
  static async getTextOCR (image) {
    if (typeof ocrNetwork === 'undefined') {
      MNUtil.showHUD("Install 'MN OCR' first")
      return undefined
    }
    try {
      let res = await ocrNetwork.OCR(image)
      MNUtil.copy(res)
      return res
    } catch (error) {
      MNUtil.showHUD("Error in getTextOCR: "+error)
      return undefined
    }
  }

  static async fetchKeys (authorization,string) {
  let url = `https://file.feliks.top/sharedKeys.json`

  try {
    const res = await this.fetch(url,
      {
        method: "Get",
        timeout: 60
      }
    )
    return res;
  } catch (error) {
    MNUtil.showHUD(error)

    return undefined
  }
  }

  static initRequest(url,options) {
    const request = this.requestWithURL(url)
    request.setHTTPMethod(options.method ?? "GET")
    request.setTimeoutInterval(options.timeout ?? 10)
    const headers = {
      "User-Agent":
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.1 Safari/605.1.15",
      "Content-Type": "application/json",
      Accept: "application/json"
    }
    request.setAllHTTPHeaderFields({
      ...headers,
      ...(options.headers ?? {})
    })
    if (options.search) {
      request.setURL(
        this.genNSURL(
          `${url.trim()}?${Object.entries(options.search).reduce((acc, cur) => {
            const [key, value] = cur
            return `${acc ? acc + "&" : ""}${key}=${encodeURIComponent(value)}`
          }, "")}`
        )
      )
    } else if (options.body) {
      request.setHTTPBody(NSData.dataWithStringEncoding(options.body, 4))
    } else if (options.form) {
      request.setHTTPBody(
        NSData.dataWithStringEncoding(
          Object.entries(options.form).reduce((acc, cur) => {
            const [key, value] = cur
            return `${acc ? acc + "&" : ""}${key}=${encodeURIComponent(value)}`
          }, ""),
          4
        )
      )
    } else if (options.json) {
      request.setHTTPBody(
        NSJSONSerialization.dataWithJSONObjectOptions(
          options.json,
          1
        )
      )
    }
    return request
  }
  static async getUsage(authorization,string) {
    let url = chatAIConfig.config.url
    let urlSubscription = url+'/v1/dashboard/billing/subscription'
    let urlUsage = url+"/v1/dashboard/billing/usage"

    let key = chatAIConfig.config.openaiKey
    if (key.trim() === "") {
      MNUtil.showHUD("no api key")
      return
    }
    let headers = {
        'Authorization': 'Bearer '+key,
        "Content-Type": "application/json"
    }
    let usage = {}
    const res = await this.fetch(urlUsage,
      {
        method: "GET",
        timeout: 60,
        headers: headers
      }
    )
    usage.usage = res.total_usage
    const total = await this.fetch(urlSubscription,
      {
        method: "GET",
        timeout: 60,
        headers: headers
      }
    )
    usage.total = total.hard_limit_usd
    return usage;
  }
/**
 * 
 * @param {*} fullPath 
 * @returns {Promise<{object:String,status:String,id:String,purpose:String,bytes:Number,filename:String}>}
 */
  static async upload(fullPath,key) {
  let fileData = chatAIUtils.getFile(fullPath)
  // let fileSizeWithMB = fileData.length()/1048576
  if (fileData.length() >= 104857600) {
    MNUtil.showHUD("Too large file!")
    return {}
  }
  // copy("length:"+fileSizeWithMB+"MB")
  // showHUD("length:"+fileData.length())
  // let fileName = (fullPath)
  let fileName = `doc_${Date.now()}.pdf`
  const headers = {
    Authorization: "Bearer "+key
  }
  let url = "https://api.moonshot.cn/v1/files"
  // copyJSON(body)
  // showHUD(fileName)
  const request = this.initFileRequest(url, {
      headers: headers,
      timeout: 60
    },
    "file-extract",
    fileName,
    fileData)
  let res = await this.sendRequest(request)
  return res
//   const queue = NSOperationQueue.mainQueue()

//   return new Promise((resolve, reject) => {
    
//   NSURLConnection.sendAsynchronousRequestQueueCompletionHandler(
//     request,
//     queue,
//     (res, data, err) => {
//       if (err.localizedDescription){
//         showHUD(err.localizedDescription)
//         reject()
//       }
//       // if (data.length() === 0) resolve({})
//       const result = NSJSONSerialization.JSONObjectWithDataOptions(
//         data,
//         1<<0
//       )
//       if (NSJSONSerialization.isValidJSONObject(result)){
//         // showHUD(result)
//         // copyJSON(result)
//         resolve(result)
//       }
//       resolve(result)
//       // showHUD(result)
//     }
//   )
//   })
}
  static getAuthorization(apikey) {
    let header = JSON.stringify({"alg":"HS256","sign_type":"SIGN"})
    let payload =JSON.stringify({
      "api_key": apikey.split(".")[0],
      "exp": Date.now()+60000,
      "timestamp": Date.now()
    });
    let secretSalt = apikey.split(".")[1];
    let before_sign = this.base64UrlEncode(CryptoJS.enc.Utf8.parse(header)) + '.' + this.base64UrlEncode(CryptoJS.enc.Utf8.parse(payload));

    let  signature =CryptoJS.HmacSHA256(before_sign, secretSalt);
     signature = this.base64UrlEncode(signature);

    let final_sign = before_sign + '.' + signature;
    return final_sign
  }
   //和普通base64加密不一样
  static base64UrlEncode(str) {
     var encodedSource = CryptoJS.enc.Base64.stringify(str);
     var reg = new RegExp('/', 'g');
     encodedSource = encodedSource.replace(/=+$/,'').replace(/\+/g,'-').replace(reg,'_');
     return encodedSource;
  }
}
